/*
 * Copyright (C) 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "aiIndex.h"

#include <kyai-business-framework/conf.h>
#include <kyai-business-framework/log.h>
#include <kyai-business-framework/rpc/vectorDB/vectorDatabase.h>

#include "fileInfoService.h"
#include "summaryTask.h"
#include "systemCallback.h"
#include "tagTask.h"
#include "textEmbeddingTask.h"

namespace DataManagement {
AIIndex& AIIndex::getInstance() {
    static AIIndex aiIndex;
    return aiIndex;
}

AIIndex::AIIndex() : idleSignalConnection(nullptr), idleSignalId(0), idleState(IDLE_STATE_BUSY) {}

int AIIndex::exec() {
    syncData();

    if (listenSignal()) {
        return -1;
    }

    createWorkTask();

    startWorkTask();

    return 0;
}

void AIIndex::setIdleState(IdleState newStatus) {
    PRINT_INFO("SetIdleState from %d to %d.\n", idleState, newStatus);

    std::unique_lock<std::shared_mutex> lock(idleLock);
    if (!checkIdleState(newStatus)) {
        PRINT_ERROR("SetIdleState status is wrong, from %d to %d.\n", idleState, newStatus);
        return;
    }

    idleState = newStatus;
}

void AIIndex::notifyAllTask() {
    for (const auto& worker : taskLists) {
        worker->awakeSelf();
    }
}

bool AIIndex::isIdle() {
    std::shared_lock<std::shared_mutex> lock(idleLock);
    return idleState == IDLE_STATE_TRUE_IDLE;
}

int AIIndex::listenSignal() {
    PRINT_INFO("Start listening system idle signal...\n");

    // 连接到会话总线，dbus demo的方式
    GError* error = nullptr;
    idleSignalConnection = g_bus_get_sync(G_BUS_TYPE_SESSION, nullptr, &error);
    if (idleSignalConnection == nullptr) {
        if (error) {
            PRINT_ERROR("Connecting dbus demo failed, error: %s.\n", error->message);
            g_error_free(error);
        } else {
            PRINT_ERROR("Connecting dbus demo failed, but error is null.\n");
        }
        return -1;
    }

    // 系统空闲信号的地址
    const char* sender = "org.gnome.SessionManager";
    const char* interfaceName = "org.gnome.SessionManager.Presence";
    const char* objectPath = "/org/gnome/SessionManager/Presence";
    const char* single = "StatusChanged";
    idleSignalId = g_dbus_connection_signal_subscribe(idleSignalConnection, sender, interfaceName, single, objectPath,
                                                      nullptr, G_DBUS_SIGNAL_FLAGS_NONE, judgeIsIdle, nullptr, nullptr);

    PRINT_INFO("Start listening system idle signal success, id: %d\n", idleSignalId);
    return 0;
}

bool AIIndex::checkIdleState(IdleState newStatus) const {
    switch (newStatus) {
        case IDLE_STATE_BUSY:
            return idleState == IDLE_STATE_SYSTEM_IDLE || idleState == IDLE_STATE_TRUE_IDLE;
        case IDLE_STATE_SYSTEM_IDLE:
            return idleState == IDLE_STATE_BUSY;
        case IDLE_STATE_TRUE_IDLE:
            return idleState == IDLE_STATE_SYSTEM_IDLE;
        default:
            return false;
    }
}

void AIIndex::createWorkTask() {
    const std::map<int, std::shared_ptr<BaseTask>> taskTypeMap = {
        {TASK_TYPE_TEXT_EMBEDDING, std::make_shared<TextEmbeddingTask>()},
        {TASK_TYPE_TAG, std::make_shared<TagTask>()},
        {TASK_TYPE_SUMMARY, std::make_shared<SummaryTask>()},
    };

    std::vector<int> taskType = Configuration::getIntArray("ai_index_task_type", {1, 2, 3});
    for (const auto& item : taskType) {
        const std::map<int, std::shared_ptr<BaseTask>>::const_iterator& it = taskTypeMap.find(item);
        if (it == taskTypeMap.end()) {
            PRINT_ERROR("createWorkTask cannot find task type: %d.\n", item);
            continue;
        }
        PRINT_INFO("createWorkTask %d.\n", item);
        taskLists.push_back(it->second);
    }
}

void AIIndex::startWorkTask() {
    for (const auto& worker : taskLists) {
        worker->start();
    }
}

/**
 * 版本升级，同步已经建好的数据
 */
void AIIndex::syncData() {
    SQLite::Database db = FileDatabase::createDatabase();
    if (!db.tableExists("Files")) {
        PRINT_INFO("AIIndex::syncData Files table not exist.\n");
        return;
    }

    SQLite::Statement query(db, "SELECT file_id, file_path FROM Files");
    int rows;
    while (query.executeStep()) {
        int64_t fileId = query.getColumn(0);
        std::string filePath = query.getColumn(1);
        std::string modifyTime = FileInfoService::getLastModifyTime(filePath);

        // 图片向量化已经做完，无需创建任务
        const std::vector<std::string> imageSupportFormat = {"png", "jpg", "jpeg", "jpe", "bmp", "dib"};
        if (std::find(imageSupportFormat.begin(), imageSupportFormat.end(), filePath.substr(filePath.rfind('.') + 1)) !=
            imageSupportFormat.end()) {
            PRINT_INFO("file path: %s, is image.\n", filePath.c_str());
            SQLite::Statement fileInfoStatement(
                db, "INSERT INTO t_file_info(file_id, enable, file_path, file_modify_time) VALUES(?, 1, ?, ?)");
            fileInfoStatement.bind(1, fileId);
            fileInfoStatement.bind(2, filePath);
            fileInfoStatement.bind(3, modifyTime);

            rows = fileInfoStatement.exec();
            PRINT_INFO("SQL: %s, success %d rows.\n", fileInfoStatement.getExpandedSQL().c_str(), rows);
            continue;
        }

        SQLite::Statement fileInfoStatement(
            db, "INSERT INTO t_file_info(file_id, file_path, file_modify_time) VALUES(?, ?, ?)");
        fileInfoStatement.bind(1, fileId);
        fileInfoStatement.bind(2, filePath);
        fileInfoStatement.bind(3, modifyTime);

        rows = fileInfoStatement.exec();
        PRINT_INFO("SQL: %s, success %d rows.\n", fileInfoStatement.getExpandedSQL().c_str(), rows);

        // 同步过来的数据，向量化已经做完，标签和摘要未做完
        SQLite::Statement aiIndexStatement(
            db, "INSERT INTO t_ai_index_task(file_id, file_path, state, type) VALUES(?,?,?,?), (?,?,?,?), (?,?,?,?)");
        aiIndexStatement.bind(1, fileId);
        aiIndexStatement.bind(2, filePath);
        aiIndexStatement.bind(3, TASK_STATE_SUCCESS);
        aiIndexStatement.bind(4, TASK_TYPE_TEXT_EMBEDDING);
        aiIndexStatement.bind(5, fileId);
        aiIndexStatement.bind(6, filePath);
        aiIndexStatement.bind(7, TASK_STATE_INIT);
        aiIndexStatement.bind(8, TASK_TYPE_TAG);
        aiIndexStatement.bind(9, fileId);
        aiIndexStatement.bind(10, filePath);
        aiIndexStatement.bind(11, TASK_STATE_INIT);
        aiIndexStatement.bind(12, TASK_TYPE_SUMMARY);
        PRINT_INFO("SQL: %s.\n", aiIndexStatement.getExpandedSQL().c_str());
        rows = aiIndexStatement.exec();
        PRINT_INFO("SQL: %s, success %d rows.\n", aiIndexStatement.getExpandedSQL().c_str(), rows);
    }

    db.exec("DROP TABLE Files");
    db.exec("DROP TABLE Tags");
    db.exec("DROP TABLE FileTags");
    PRINT_INFO("AIIndex::syncData drop table: Files, Tags, FileTags.\n");

    scene::VectorDatabase::getInstance().dropCollection("files_tags");
    PRINT_INFO("AIIndex::syncData drop vector-db collection: files_tags.\n");
}

AIIndex::~AIIndex() {
    if (idleSignalConnection && idleSignalId > 0) {
        g_dbus_connection_signal_unsubscribe(idleSignalConnection, idleSignalId);
        idleSignalId = 0;
        g_object_unref(idleSignalConnection);
        idleSignalConnection = nullptr;
    }
}
}  // namespace DataManagement
